from copy import deepcopy

import geopandas as gpd
import iris
import numpy as np
import xarray as xr

from CoCoMET.analysis.analysis_object import Analysis_Object
from CoCoMET.analysis.calc_var import calc_var
from CoCoMET.post_processor import filter_cells

# Observations
from .goes_load import goes_load_netcdf_iris
from .goes_tams import goes_run_tams
from .goes_tobac import (
    goes_tobac_feature_id,
    goes_tobac_linking,
    goes_tobac_segmentation,
)
from .mesonh_calculate_products import mesonh_calculate_reflectivity
from .mesonh_load import mesonh_load_netcdf_iris
from .mesonh_moaap import mesonh_run_moaap
from .mesonh_tams import mesonh_run_tams
from .mesonh_tobac import (
    mesonh_tobac_feature_id,
    mesonh_tobac_linking,
    mesonh_tobac_segmentation,
)
from .multi_nexrad_load import multi_nexrad_load_netcdf_iris
from .multi_nexrad_tobac import (
    multi_nexrad_tobac_feature_id,
    multi_nexrad_tobac_linking,
    multi_nexrad_tobac_segmentation,
)
from .nexrad_load import nexrad_load_netcdf_iris
from .nexrad_tobac import (
    nexrad_tobac_feature_id,
    nexrad_tobac_linking,
    nexrad_tobac_segmentation,
)
from .rams_calculate_products import rams_calculate_reflectivity
from .rams_load import rams_load_netcdf_iris
from .rams_moaap import rams_run_moaap
from .rams_tams import rams_run_tams
from .rams_tobac import (
    rams_tobac_feature_id,
    rams_tobac_linking,
    rams_tobac_segmentation,
)
from .standard_radar_load import standard_radar_load_netcdf_iris
from .standard_radar_tobac import (
    standard_radar_tobac_feature_id,
    standard_radar_tobac_linking,
    standard_radar_tobac_segmentation,
)

# Load the US converting functions
from .tracker_output_translation_layer import (
    bulk_moaap_to_US,
    bulk_tams_to_US,
    feature_id_to_US,
    linking_to_US,
    segmentation_to_US,
)

# Models
from .wrf_calculate_products import wrf_calculate_reflectivity
from .wrf_load import wrf_load_netcdf_iris
from .wrf_moaap import wrf_run_moaap
from .wrf_tams import wrf_run_tams
from .wrf_tobac import wrf_tobac_feature_id, wrf_tobac_linking, wrf_tobac_segmentation

# Loading these is not memory or time intensive and is best practice to do so in the file header.


################################################################
#################### RUN TRACKING PROGRAMS #####################
################################################################


def _create_xarrs_and_cubes(
    dataset_name: str, CONFIG: dict
) -> tuple[xr.Dataset, iris.cube, xr.Dataset, iris.cube]:
    """


    Parameters
    ----------
    dataset_name : str
        The name of the dataset.
    CONFIG : dict
        User configuration file.

    Returns
    -------
    tracking_xarray : xr.Dataset
        An xarray with the feature tracking information.
    tracking_cube : iris.cube
        An iris cube of the feature tracking variable
    segmentation_xarray : xr.Dataset
        An xarray with the feature segmentation information
    segmentation_cube : iris.cube
        An iris cube of the feature segmentation variables
    """

    # Load in the datasets
    # some datasets have different load_netcdf arguments, so account for the differences
    # TODO: eventually we will call RAMS-MAT in rams_load_netcdf_iris, so the only exceptions will be the nexrads
    if dataset_name == "rams":
        tracking_cube, tracking_xarray = globals()[f"{dataset_name}_load_netcdf_iris"](
            CONFIG[dataset_name]["path_to_data"],
            CONFIG[dataset_name]["feature_tracking_var"],
            CONFIG[dataset_name]["path_to_header"],
            CONFIG,
        )

    elif dataset_name == "nexrad" or dataset_name == "multi_nexrad":
        if "gridding" in CONFIG[dataset_name]:
            if CONFIG["verbose"]:
                print(f"=====Gridding {dataset_name.upper()} Data=====")
            tracking_cube, tracking_xarray = globals()[
                f"{dataset_name}_load_netcdf_iris"
            ](
                CONFIG[dataset_name]["path_to_data"],
                "ar2v",
                CONFIG[dataset_name]["feature_tracking_var"],
                CONFIG,
                CONFIG[dataset_name]["gridding"]["gridding_save_path"],
            )

        else:
            tracking_cube, tracking_xarray = globals()[
                f"{dataset_name}_load_netcdf_iris"
            ](
                CONFIG[dataset_name]["path_to_data"],
                "nc",
                CONFIG[dataset_name]["feature_tracking_var"],
                CONFIG,
            )

    else:
        tracking_cube, tracking_xarray = globals()[f"{dataset_name}_load_netcdf_iris"](
            CONFIG[dataset_name]["path_to_data"],
            CONFIG[dataset_name]["feature_tracking_var"],
            CONFIG,
        )

    # if tracking and segmentation variables are different, load seperately
    if (
        CONFIG[dataset_name]["feature_tracking_var"]
        != CONFIG[dataset_name]["segmentation_var"]
    ):
        if dataset_name == "rams":
            segmentation_cube, segmentation_xarray = globals()[
                f"{dataset_name}_load_netcdf_iris"
            ](
                CONFIG[dataset_name]["path_to_data"],
                CONFIG[dataset_name]["segmentation_var"],
                CONFIG[dataset_name]["path_to_header"],
                CONFIG,
            )
        # TODO: check if you need to copy and paste the nexrad/multi-nexrad load configurations for segmentation
        else:
            tracking_cube, tracking_xarray = globals()[
                f"{dataset_name}_load_netcdf_iris"
            ](
                CONFIG[dataset_name]["path_to_data"],
                CONFIG[dataset_name]["feature_tracking_var"],
                CONFIG,
            )

    else:
        segmentation_cube = tracking_cube
        segmentation_xarray = tracking_xarray

    # if reflectivity is required for analysis, add it now
    # currently only implemented for tobac (first clause)
    if ("tobac" in CONFIG[dataset_name]) and (
        "analysis" in CONFIG[dataset_name]["tobac"]
    ):
        if (
            (
                dataset_name == "rams"
                or dataset_name == "wrf"
                or dataset_name == "mesonh"
            )
            and ("eth" in CONFIG[dataset_name]["tobac"]["analysis"])
            and CONFIG[dataset_name]["segmentation_var"].lower() != "dbz"
        ):
            reflectivity_calc = globals()[f"{dataset_name}_calculate_reflectivity"](
                tracking_xarray
            )
            tracking_xarray["DBZ"] = reflectivity_calc
            segmentation_xarray["DBZ"] = reflectivity_calc

    return (tracking_xarray, tracking_cube, segmentation_xarray, segmentation_cube)


def _run_tracker_det_and_seg(
    dataset_name: str,
    tracker: str,
    tracking_info: tuple[xr.Dataset, iris.cube, xr.Dataset, iris.cube],
    CONFIG: dict,
) -> tuple[
    xr.Dataset,
    xr.Dataset,
    list[gpd.GeoDataFrame, gpd.GeoDataFrame, xr.Dataset, xr.Dataset],
]:
    """


    Run feature detection and segmentation on a dataset for a given tracker

    Parameters
    ----------
    dataset_name : str
        The name of the dataset.
    tracker : str
        The name of the tracker.
    tracking_info : tuple[xr.Dataset, iris.cube, xr.Dataset, iris.cube]
        A tuple of the feature/segmentation xarrays/iris cubes.
    CONFIG : dict
        User configuration file.

    Returns
    -------
    tracking_xarray : xr.Dataset
        An xarray with the feature tracking information.
    segmentation_xarray : xr.Dataset, returned only if tracker == "tobac"
        An xarray with the feature segmentation information
    US_values : list[gpd.GeoDataFrame, gpd.GeoDataFrame, xr.Dataset, xr.Dataset]
        US_features : gpd.GeoDataFrame
            A CoCoMET-US standard features GeoDataFrame.
        US_tracks : gpd.GeoDataFrame
            A CoCoMET-US standard trakcs GeoDataFrame.
        US_segmentation_2d : xr.Dataset
            A CoCoMET-US standard 2D segmentation xarray.
        US_segmentation_3d : xr.Dataset
            A CoCoMET-US standard 3D segmentation xarray.
    """
    # TODO : create error warnings in case some datasets do not have some trackers implemented yet

    # Load in tracking and segmentation info
    tracking_xarray = tracking_info[0]
    tracking_cube = tracking_info[1]
    segmentation_xarray = tracking_info[2]
    segmentation_cube = tracking_info[3]

    # make sure the tracker name is in lower case
    tracker = tracker.lower()

    # now determine which tracker(s) to use
    if tracker == "tobac":
        features = None
        tracks = None
        segmentation2d = (None, None)
        segmentation3d = (None, None)

        # Perform all cell tracking, id, and segmentation steps. Then add results to return dict
        if "feature_id" in CONFIG[dataset_name]["tobac"]:
            if CONFIG["verbose"]:
                print(f"=====Starting {dataset_name.upper()} tobac Feature ID=====")

            features = globals()[f"{dataset_name}_tobac_feature_id"](
                tracking_cube, CONFIG
            )
            if features is None:
                raise Exception("No features identified")

        if "linking" in CONFIG[dataset_name]["tobac"]:
            if CONFIG["verbose"]:
                print(
                    f"=====Starting {dataset_name.upper()} tobac Feature Linking====="
                )

            tracks = globals()[f"{dataset_name}_tobac_linking"](
                tracking_cube, features, CONFIG
            )

        if "segmentation_2d" in CONFIG[dataset_name]["tobac"]:
            if CONFIG["verbose"]:
                print(
                    f"=====Starting {dataset_name.upper()} tobac 2D Segmentation====="
                )

            height = (
                CONFIG[dataset_name]["tobac"]["segmentation_2d"]["height"]
                if "height" in CONFIG[dataset_name]["tobac"]["segmentation_2d"]
                else None
            )

            segmentation2d = globals()[f"{dataset_name}_tobac_segmentation"](
                segmentation_cube,
                features,
                "2d",
                CONFIG,
                height,
            )

        if "segmentation_3d" in CONFIG[dataset_name]["tobac"]:
            if CONFIG["verbose"]:
                print(
                    f"=====Starting {dataset_name.upper()} tobac 3D Segmentation====="
                )

            segmentation3d = globals()[f"{dataset_name}_tobac_segmentation"](
                segmentation_cube, features, "3d", CONFIG
            )

        # Create analysis object values
        US_features = feature_id_to_US(features, "tobac")
        US_tracks = linking_to_US(tracks, "tobac")
        US_segmentation_2d = segmentation_to_US(segmentation2d[0], US_tracks, "tobac")
        US_segmentation_3d = segmentation_to_US(segmentation3d[0], US_tracks, "tobac")

        US_values = [
            US_features,
            US_tracks,
            US_segmentation_2d,
            US_segmentation_3d,
        ]
        return (tracking_xarray, segmentation_xarray, US_values)

    if tracker == "moaap" or tracker == "tams":
        # Run MOAAP or TAMS if present
        if tracker in CONFIG[dataset_name]:
            if CONFIG["verbose"]:
                print(
                    f"=====Starting {dataset_name.upper()} {tracker.upper()} Tracking====="
                )

            # Run MOAAP or TAMS
            mask_output = globals()[f"{dataset_name}_run_{tracker}"](
                tracking_xarray, CONFIG
            )

            try:  # TODO: the models and observations have different naming conventions for their projection coordinates
                projection_x = tracking_xarray.PROJX.values
                projection_y = tracking_xarray.PROJY.values
            except:
                projection_x = tracking_xarray.projection_x_coordinate.values
                projection_y = tracking_xarray.projection_y_coordinate.values

            # Convert output to US and check if is None
            if tracker == "tams":
                tams_output, latlon_coord_system = mask_output
                if tams_output is None:
                    US_values = [None, None, None]

                else:
                    US_values = bulk_tams_to_US(
                        tams_output,
                        latlon_coord_system,
                        projection_x,
                        projection_y,
                        convert_type=CONFIG[dataset_name]["tams"]["analysis_type"],
                    )

            elif tracker == "moaap":
                US_values = bulk_moaap_to_US(
                    mask_output,
                    projection_x,
                    projection_y,
                    convert_type=CONFIG[dataset_name][tracker]["analysis_type"],
                )

                if US_values is None:
                    US_values = [None, None, None]

        return (tracking_xarray, US_values)


################################################################
#################### RUN ANALYSIS PROGRAMS #####################
################################################################


def _tobac_analysis(
    dataset_name: str,
    user_return_dict: dict,
    analysis_object: Analysis_Object,
    CONFIG: dict,
) -> dict:
    """


    Parameters
    ----------
    dataset_name : str
        The name of the dataset.
    user_return_dict : dict
        A dictionary that is appended to which contains all of the analysis and tracking information for a dataset.
    analysis_object : Analysis_Object
        A  CoCoMET-US standard analysis object containing at least US_tracks, US_segmentation_2d, and US_segmentation_3d.
    CONFIG : dict
        User configuration file.

    Returns
    -------
    user_return_dict : dict
        A dictionary that is appended to which contains all of the analysis and tracking information for a dataset.
    """

    analysis_dictionary = analysis_object.return_analysis_dictionary()
    US_tracks = analysis_dictionary["US_tracks"]
    US_segmentation_2d = analysis_dictionary["US_segmentation_2d"]
    US_segmentation_3d = analysis_dictionary["US_segmentation_3d"]

    # Make an empty dictionary to fill with analysis data
    _tobac_analysis_data = {}

    if "analysis" in CONFIG[dataset_name]["tobac"]:
        if CONFIG[dataset_name]["tobac"]["analysis"] is None:
            CONFIG[dataset_name]["tobac"]["analysis"] = {}

        if CONFIG["verbose"]:
            print(
                f"=====Starting {dataset_name.upper()} tobac Analysis Calculations====="
            )

        # Place the analysis variables which help calculate other variables first in the list if they are to be calculated
        analysis_dict = CONFIG[dataset_name]["tobac"]["analysis"]
        analysis_keys = list(analysis_dict.keys())
        depended_variables = ["volume", "perimeter", "area"]

        for dep_var in depended_variables:
            if dep_var in analysis_keys:
                analysis_keys.remove(dep_var)
                analysis_keys.insert(0, dep_var)

        analysis_vars = {key: analysis_dict[key] for key in analysis_keys}

        # Calcaulte each variable of interest and append to analysis data array
        for var in analysis_vars:

            # Add default tracking featured_id variable in place of variable if not present
            if "variable" not in CONFIG[dataset_name]["tobac"]["analysis"][var]:

                CONFIG[dataset_name]["tobac"]["analysis"][var]["variable"] = CONFIG[
                    dataset_name
                ]["feature_tracking_var"].upper()

            # This allows us to have multiple copies of the same variable by adjoining a dash
            proper_var_name = var.lower().split("-")[0]

            arg_dictionary = deepcopy(_tobac_analysis_data)
            arg_dictionary.update(CONFIG[dataset_name]["tobac"]["analysis"][var])
            _tobac_analysis_data[var] = calc_var(
                analysis_object,
                proper_var_name,
                **arg_dictionary,
            )

    if CONFIG["verbose"]:
        print(f"=====Converting {dataset_name.upper()} tobac Output to CoCoMET-US=====")

    # Add all products to return dict
    user_return_dict[dataset_name]["tobac"] = {
        "US_tracks": US_tracks,
        "US_segmentation_2d": US_segmentation_2d,
        "US_segmentation_3d": US_segmentation_3d,
        "analysis": _tobac_analysis_data,
        "analysis_object": analysis_object,
    }

    if CONFIG["verbose"]:
        print(f"====={dataset_name.upper()} tobac Tracking Complete=====")

    return user_return_dict


def _moaap_analysis(
    dataset_name: str,
    user_return_dict: dict,
    analysis_object: Analysis_Object,
    CONFIG: dict,
) -> dict:
    """


    Parameters
    ----------
    dataset_name : str
        The name of the dataset.
    user_return_dict : dict
        A dictionary that is appended to which contains all of the analysis and tracking information for a dataset.
    analysis_object : Analysis_Object
        A  CoCoMET-US standard analysis object containing at least US_tracks, US_segmentation_2d, and US_segmentation_3d.
    CONFIG : dict
        User configuration file.

    Returns
    -------
    user_return_dict : dict
        A dictionary that is appended to which contains all of the analysis and tracking information for a dataset.
    """
    analysis_dictionary = analysis_object.return_analysis_dictionary()
    US_tracks = analysis_dictionary["US_tracks"]
    US_segmentation_2d = analysis_dictionary["US_segmentation_2d"]
    US_segmentation_3d = analysis_dictionary["US_segmentation_3d"]

    # Make an empty dictionary to fill with analysis data
    _moaap_analysis_data = {}

    # Run analysis on MOAAP output
    if "analysis" in CONFIG[dataset_name]["moaap"]:
        if CONFIG[dataset_name]["moaap"]["analysis"] is None:
            CONFIG[dataset_name]["moaap"]["analysis"] = {}

        if CONFIG["verbose"]:
            print(
                f"=====Starting {dataset_name.upper()} MOAAP Analysis Calculations====="
            )

        if (US_tracks is not None and US_segmentation_2d is not None) or (
            US_tracks is not None and US_segmentation_3d is not None
        ):
            # Calcaulte each variable of interest and append to analysis data array
            for var in CONFIG[dataset_name]["moaap"]["analysis"].keys():
                # Add default tracking featured_id variable in place of variable if not present
                if "variable" not in CONFIG[dataset_name]["moaap"]["analysis"][var]:
                    CONFIG[dataset_name]["moaap"]["analysis"][var]["variable"] = CONFIG[
                        dataset_name
                    ]["feature_tracking_var"].upper()

                # This allows us to have multiple copies of the same variable by adjoining a dash
                proper_var_name = var.lower().split("-")[0]

                _moaap_analysis_data[var] = calc_var(
                    analysis_object,
                    proper_var_name,
                    **CONFIG[dataset_name]["moaap"]["analysis"][var],
                )
        else:
            print(
                f"=====No MOAAP Tracking Information Found for {dataset_name}, Skipping Analysis====="
            )

    user_return_dict[dataset_name]["moaap"] = {
        # "mask_xarray": mask_output,
        # "US_features": US_values[0],
        "US_tracks": US_tracks,
        "US_segmentation_2d": US_segmentation_2d,
        "analysis": _moaap_analysis_data,
        "analysis_object": analysis_object,
    }

    if CONFIG["verbose"]:
        print(f"====={dataset_name.upper()} MOAAP Tracking Complete=====")

    return user_return_dict


def _tams_analysis(
    dataset_name: str,
    user_return_dict: dict,
    analysis_object: Analysis_Object,
    CONFIG: dict,
) -> dict:
    """


    Parameters
    ----------
    dataset_name : str
        The name of the dataset.
    user_return_dict : dict
        A dictionary that is appended to which contains all of the analysis and tracking information for a dataset.
    analysis_object : Analysis_Object
        A  CoCoMET-US standard analysis object containing at least US_tracks, US_segmentation_2d, and US_segmentation_3d.
    CONFIG : dict
        User configuration file.

    Returns
    -------
    user_return_dict : dict
        A dictionary that is appended to which contains all of the analysis and tracking information for a dataset.
    """

    analysis_dictionary = analysis_object.return_analysis_dictionary()
    US_tracks = analysis_dictionary["US_tracks"]
    US_segmentation_2d = analysis_dictionary["US_segmentation_2d"]
    US_segmentation_3d = analysis_dictionary["US_segmentation_3d"]

    # Make an empty dictionary to fill with analysis data
    _tams_analysis_data = {}

    # Run analysis on TAMS output
    if "analysis" in CONFIG[dataset_name]["tams"]:
        if CONFIG[dataset_name]["tams"]["analysis"] is None:
            CONFIG[dataset_name]["tams"]["analysis"] = {}

        if CONFIG["verbose"]:
            print(
                f"=====Starting {dataset_name.upper()} TAMS Analysis Calculations====="
            )

        if (US_tracks is not None and US_segmentation_2d is not None) or (
            US_tracks is not None and US_segmentation_3d is not None
        ):
            # Calcaulte each variable of interest and append to analysis data array
            for var in CONFIG[dataset_name]["tams"]["analysis"].keys():
                # Add default tracking featured_id variable in place of variable if not present
                if "variable" not in CONFIG[dataset_name]["tams"]["analysis"][var]:
                    CONFIG[dataset_name]["tams"]["analysis"][var]["variable"] = CONFIG[
                        dataset_name
                    ]["feature_tracking_var"].upper()

                # This allows us to have multiple copies of the same variable by adjoining a dash
                proper_var_name = var.lower().split("-")[0]

                _tams_analysis_data[var] = calc_var(
                    analysis_object,
                    proper_var_name,
                    **CONFIG[dataset_name]["tams"]["analysis"][var],
                )
        else:
            print(
                f"=====No TAMS Tracking Information Found for {dataset_name}, Skipping Analysis====="
            )

    user_return_dict[dataset_name]["tams"] = {
        # "mask_xarray": mask_output,
        # "US_features": US_values[0],
        "US_tracks": US_tracks,
        "US_segmentation_2d": US_segmentation_2d,
        "analysis": _tams_analysis_data,
        "analysis_object": analysis_object,
    }

    if CONFIG["verbose"]:
        print(f"====={dataset_name.upper()} TAMS Tracking Complete=====")

    return user_return_dict


################################################################
####################### RUN ALL PROGRAMS #######################
################################################################


# run_tracker needs to be at the bottom of the script to access _moaap_analysis and _tams_analysis
def run_tracker(dataset_name, tracker, user_return_dict, tracking_info, CONFIG):
    """
    Run a tracker on a dataset. If no tracker has been previously run on the dataset,
    user_return_dict should be empty

    Inputs:
        dataset_name : a string giving the name of the data you would like to track
        tracker : a string giving the name of the tracker you would like to use
        user_return_dict : a dictionary of the previous dataset processed information
        CONFIG : an ANTE-TRACE CONFIG file

    Returns:
        user_return_dict : a dictionary with current processed information
    """

    # Create a user dictionary key for the dataset if there is not one
    if dataset_name not in user_return_dict:
        user_return_dict[dataset_name] = {}

    # If the iris cubes have not been created already, do so now
    if tracking_info is None:
        tracking_info = _create_xarrs_and_cubes(dataset_name, CONFIG)

    # Determine which trackers to track on

    if tracker == "tobac":
        (tracking_xarray, segmentation_xarray, US_values) = _run_tracker_det_and_seg(
            dataset_name, tracker, tracking_info, CONFIG
        )

        # Create analysis object - for tobac
        analysis_object = Analysis_Object(
            tracking_xarray, segmentation_xarray, *US_values
        )

        # Filter the cells
        # analysis_object = filter_cells(analysis_object) # there is an issue with the memory for high resolution data sets

        user_return_dict = _tobac_analysis(
            dataset_name, user_return_dict, analysis_object, CONFIG
        )

        return user_return_dict, tracking_info

    if tracker == "moaap" or tracker == "tams":
        tracking_xarray, US_values = _run_tracker_det_and_seg(
            dataset_name, tracker, tracking_info, CONFIG
        )

        # Create analysis object - for moaap
        analysis_object = Analysis_Object(
            tracking_xarray,
            tracking_xarray,
            *US_values,
            None,  # No 3d segmentation
        )

        # Filter the cells
        # analysis_object = filter_cells(analysis_object)

        if tracker == "moaap":
            user_return_dict = _moaap_analysis(
                dataset_name,
                user_return_dict,
                analysis_object,
                CONFIG,
            )

        elif tracker == "tams":
            user_return_dict = _tams_analysis(
                dataset_name,
                user_return_dict,
                analysis_object,
                CONFIG,
            )

        return user_return_dict, tracking_info



from copy import deepcopy

import numpy as np
import pandas as pd
import scipy.spatial as sp
from skimage import measure
from tqdm import tqdm

from .projection_calc_3d import point_projection


def convexity(
    analysis_dict: dict,
    surface_area_df: pd.DataFrame,  # What is this?
    segmentation_type: str = None,
) -> pd.DataFrame:
    """


    Parameters
    ----------
    analysis_dict : dict
        Dictionary output from analysis_object.return_analysis_dictionary(). Contains US_tracks dataframe and segmentation xarray dataset..
    surface_area_df : pd.DataFrame
        ???.
    segmentation_type : str, optional
        Whether to calculate 2d or 3d convexity if convexity in irregularity_matrics. If None, will use 3d segmentation if it is available, or 2d segmentation if 3d segmentation is unavailable. The default is None.

    Raises
    ------
    ValueError
        No segmentation available in input dictionary or invalid segmentation_type.

    Returns
    -------
    convexity_df : pandas.core.frame.DataFrame
        Dataframe with columns of feature id and convexity, where convexity is a value between 0 and 1.

    """

    # Returns a value between 0 and 1 which represents the convexity of the cell. Lower convexity represents more irregular edges
    tracks = analysis_dict["US_tracks"]
    tracks_and_perim = tracks.join(surface_area_df)

    if segmentation_type is None:
        if analysis_dict["US_segmentation_3d"] is not None:
            segmentation_type = "3d"
        elif analysis_dict["US_segmentation_2d"] is not None:
            segmentation_type = "2d"
        else:
            raise ValueError("!=====Missing Segmentation Input=====!")

    # If 3D segmentation is available, use that
    if (
        segmentation_type.lower() == "3d"
    ):
        footprint = analysis_dict["US_segmentation_3d"].Feature_Segmentation
        convexities = []
        feature_ids = []
        frame_numbers = []
        cell_ids = []
        frame_groups = tracks_and_perim.groupby("frame")
        for frame_number, frame_df in tqdm(
            frame_groups,
            desc="=====Calculating Cell Convexities=====",
            total=len(frame_groups),
        ):
            final_mask = deepcopy(footprint[frame_number])
            data = final_mask.values
            for (
                row
            ) in (
                frame_df.itertuples()
            ):  # using itertuples because internet told me it"s faster
                id_num = row[5]
                cell_id = row[6]
                # i = row[0]
                perim = row[-1]

                feature_ids.append(id_num)
                frame_numbers.append(frame_number)
                cell_ids.append(cell_id)

                points = np.asarray(list(zip(*np.where(data == id_num))))
                if len(points) == 0:
                    convexities.append(np.NaN)
                    continue

                points_proj = []
                for point in points:  # convert to projection coordinates
                    point_proj = np.array(point_projection(footprint, point))
                    points_proj.append(point_proj / 1000)
                points_proj = np.asarray(points_proj)

                try:
                    convex_hull = sp.ConvexHull(
                        points=points_proj, #qhull_options='QJ Pp'
                    )  # generate convex hull using segmentation coords
                except:
                    convexities.append(np.NaN)
                    continue

                convex_perim = convex_hull.area  # extract surface area

                convexity_val = convex_perim / perim if perim != 0 else np.nan
                if convexity_val > 1:
                    convexities.append(1)
                    continue
                convexities.append(convexity_val)

        convexity_df = pd.DataFrame(
            data={
                "frame": frame_numbers,
                "feature_id": feature_ids,
                "cell_id": cell_ids,
                "convexity": convexities,
            }
        )
        return convexity_df

    # handle 2D case
    if (
        segmentation_type.lower() == "2d"
    ):
        footprint_data = analysis_dict["US_segmentation_2d"].Feature_Segmentation
        frame_groups = tracks.groupby("frame")

        convexities = []
        feature_ids = []
        frame_numbers = []
        cell_ids = []
        for frame_number, frame_df in tqdm(
            frame_groups,
            desc="=====Calculating Cell Convexities=====",
            total=len(frame_groups),
        ):
            final_mask = deepcopy(footprint_data[frame_number])
            data = final_mask.values
            for row in frame_df.itertuples():
                id_num = row[5]
                cell_id = row[6]
                perim = row[-1]

                feature_ids.append(id_num)
                frame_numbers.append(frame_number)
                cell_ids.append(cell_id)

                mask = data == id_num

                points = np.asarray(
                    list(zip(*np.where(data == id_num)))
                )  # coordinates to every point apart of the specified cell

                try:
                    convex_hull = sp.ConvexHull(
                        points=points, #qhull_options='QJ Pp'
                    )  # generate convex hull using segmentation coords
                    convex_perim = convex_hull.area  # extract perimeter
                except:
                    convexities.append(
                        np.NaN
                    )  # this is mainly for cells that are only along a plane, the convex hull doesn"t work for them
                    continue

                contours = measure.find_contours(
                    mask, 0.5
                )  # create a contour around the 2d shape

                if contours == []:
                    convexities.append(
                        np.NaN
                    )  # When contours can't be found, append NaN
                    continue

                # this block is used to calculate 2d perimeter but you can replace it with another calculation if you want
                prev_vertex = [-1, -1]
                perim = 0
                for vertex in contours[
                    0
                ]:  # loop through every contour coord to add up the total perimeter
                    if -1 in prev_vertex:
                        prev_vertex = vertex
                        continue
                    dist = np.linalg.norm(
                        vertex - prev_vertex
                    )  # distance between current vertex and the previous
                    perim += dist
                    prev_vertex = vertex
                final_dist = np.linalg.norm(
                    contours[0][0] - contours[0][-1]
                )  # add final part of perimeter
                perim += final_dist

                convexity_val = convex_perim / perim if perim != 0 else np.nan
                if convexity_val > 1:
                    convexities.append(1)
                    continue
                convexities.append(convexity_val)

        convexity_df = pd.DataFrame(
            data={
                "frame": frame_numbers,
                "feature_id": feature_ids,
                "cell_id": cell_ids,
                "convexity": convexities,
            }
        )
        return convexity_df

    if segmentation_type.lower() != "3d" and segmentation_type.lower() != "2d":
        raise ValueError("!=====Invalid Segmenation Type=====!")


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thur Jun 13 11:33:16 2024

@author: thahn
"""

# =============================================================================
# Loads in and converts GOES satellite data to an iris cube and/or an xarray DataArray
# =============================================================================

import warnings

import iris
import iris.cube
import numpy as np
import xarray as xr


# Used to calculate lat and lon values for GOES data which only has proj_x and proj_y values
def calc_latlon(ds: xr.Dataset) -> xr.Dataset:
    """


    Parameters
    ----------
    ds : xr.Dataset
        xarray Dataset of the GOES data.

    Returns
    -------
    ds : TYPE
        xarray Dataset of the GOES data with added lat/lon data.

    """

    # The math for this function was taken from
    # https://makersportal.com/blog/2018/11/25/goes-r-satellite-latitude-and-longitude-grid-projection-algorithm
    x = ds.x
    y = ds.y
    goes_imager_projection = ds.goes_imager_projection

    x, y = np.meshgrid(x, y)

    r_eq = goes_imager_projection.attrs["semi_major_axis"]
    r_pol = goes_imager_projection.attrs["semi_minor_axis"]
    l_0 = goes_imager_projection.attrs["longitude_of_projection_origin"] * (np.pi / 180)
    h_sat = goes_imager_projection.attrs["perspective_point_height"]
    H = r_eq + h_sat

    a = np.sin(x) ** 2 + (
        np.cos(x) ** 2 * (np.cos(y) ** 2 + (r_eq**2 / r_pol**2) * np.sin(y) ** 2)
    )
    b = -2 * H * np.cos(x) * np.cos(y)
    c = H**2 - r_eq**2

    r_s = (-b - np.sqrt(b**2 - 4 * a * c)) / (2 * a)

    s_x = r_s * np.cos(x) * np.cos(y)
    s_y = -r_s * np.sin(x)
    s_z = r_s * np.cos(x) * np.sin(y)

    lat = np.arctan(
        (r_eq**2 / r_pol**2) * (s_z / np.sqrt((H - s_x) ** 2 + s_y**2))
    ) * (180 / np.pi)
    lon = (l_0 - np.arctan(s_y / (H - s_x))) * (180 / np.pi)

    ds = ds.assign_coords({"lat": (["y", "x"], lat), "lon": (["y", "x"], lon)})
    ds.lat.attrs["units"] = "degrees_north"
    ds.lon.attrs["units"] = "degrees_east"

    return ds


def goes_load_netcdf_iris(
    path_to_files: str, tracking_var: str, CONFIG: dict
) -> tuple[iris.cube.Cube, xr.DataArray]:
    """


    Parameters
    ----------
    path_to_files : str
        Glob path to input files i.e. "/data/usr/GOES/*.nc".
    tracking_var : str
        ["tb"], variable which is going to be used for tracking--brightness temperature..
    CONFIG : dict
        User configuration file.

    Raises
    ------
    Exception
        Exception if missing GOES field on CONFIG or invalid tracking variable entered.

    Returns
    -------
    cube : iris.cube.Cube
        Iris cube continaing birhtness temperature ready for tobac tracking.
    goes_xarray : xarray.core.dataarray.DataArray
        Xarray DataArray containing GOES brightness temperature data.

    """

    # Convert to iris cube and return
    if tracking_var.lower() == "tb":
        # Import GOES satetellite data as xarray dataset
        goes_xarray = xr.open_mfdataset(
            path_to_files, coords="all", concat_dim="t", combine="nested"
        )

        # Add lat and lon to goes_xarray
        # Ignore boundary warnings
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            goes_xarray = calc_latlon(goes_xarray)

        # Get GOES spatial resolution
        spacial_res = goes_xarray.spatial_resolution

        # Convert x and y projections to x and y projection in meters according to http://meteothink.org/examples/meteoinfolab/satellite/geos-16.html
        sat_height = goes_xarray.goes_imager_projection.attrs[
            "perspective_point_height"
        ]

        # Subset location of interest
        if "goes" in CONFIG:
            # Subset time based on user inputs
            if (
                "min_frame_index" in CONFIG["goes"]
                or "max_frame_index" in CONFIG["goes"]
            ):
                min_frame = (
                    CONFIG["goes"]["min_frame_index"]
                    if "min_frame_index" in CONFIG["goes"]
                    else 0
                )
                max_frame = (
                    CONFIG["goes"]["max_frame_index"] + 1
                    if "max_frame_index" in CONFIG["goes"]
                    else goes_xarray.dims["t"]
                )

                goes_xarray = goes_xarray.isel(
                    t=np.arange(min_frame, max_frame),
                    drop=True,
                )

            if "bounds" in CONFIG["goes"]:
                mask_lon = (goes_xarray.lon >= CONFIG["goes"]["bounds"][0]) & (
                    goes_xarray.lon <= CONFIG["goes"]["bounds"][1]
                )
                mask_lat = (goes_xarray.lat >= CONFIG["goes"]["bounds"][2]) & (
                    goes_xarray.lat <= CONFIG["goes"]["bounds"][3]
                )

                goes_xarray = goes_xarray.CMI.where(mask_lon & mask_lat, drop=True)

            else:
                goes_xarray = goes_xarray.CMI

        else:
            raise Exception("""!=====CONFIG Missing "goes" Field=====!""")

        # Replace time dimension with minutes since first time and add other x y z coords
        first_time = goes_xarray.t.values[0]
        goes_xarray = goes_xarray.assign_coords(
            t=(goes_xarray.t.values - first_time) / np.timedelta64(1, "m"),
            projection_x_coordinate=("x", goes_xarray.x.values * sat_height),
            projection_y_coordinate=("y", goes_xarray.y.values * sat_height),
            south_north=("y", np.arange(goes_xarray.shape[1])),
            west_east=("x", np.arange(goes_xarray.shape[2])),
            x=("x", np.arange(goes_xarray.shape[2])),
            y=("y", np.arange(goes_xarray.shape[1])),
        )

        # Drop incorrect projection vars
        goes_xarray = goes_xarray.drop_vars(["x_image", "y_image"])

        # Add spatial resolution info
        goes_xarray = goes_xarray.assign_attrs({"spatial_resolution": spacial_res})

        # Adjust dimension names to be standards accepted by iris
        goes_xarray["t"] = goes_xarray.t.assign_attrs(
            {
                "standard_name": "time",
                "long_name": f"minutes since {first_time}",
                "units": f"minutes since {first_time}",
            }
        )
        goes_xarray["lat"] = goes_xarray.lat.assign_attrs({"standard_name": "latitude"})
        goes_xarray["lon"] = goes_xarray.lon.assign_attrs(
            {"standard_name": "longitude"}
        )
        goes_xarray["projection_x_coordinate"] = (
            goes_xarray.projection_x_coordinate.assign_attrs({"units": "m"})
        )
        goes_xarray["projection_y_coordinate"] = (
            goes_xarray.projection_y_coordinate.assign_attrs({"units": "m"})
        )

        return (goes_xarray.to_iris(), goes_xarray)

    raise Exception(
        f"!=====Invalid Tracking Variable. You Entered: {tracking_var.lower()}=====!"
    )


def goes_load_netcdf(
    path_to_files: str, tracking_var: str, CONFIG: dict
) -> xr.DataArray:
    """


    Parameters
    ----------
    path_to_files : str
        Glob path to input files i.e. "/data/usr/GOES/*.nc".
    tracking_var : str
        ["tb"], variable which is going to be used for tracking--brightness temperature..
    CONFIG : dict
        User configuration file.

    Raises
    ------
    Exception
        Exception if missing GOES field on CONFIG or invalid tracking variable entered.

    Returns
    -------
    goes_xarray : xarray.core.dataarray.DataArray
        Xarray DataArray containing GOES brightness temperature data.

    """

    # Convert to iris cube and return
    if tracking_var.lower() == "tb":
        # Import GOES satetellite data as xarray dataset
        goes_xarray = xr.open_mfdataset(
            path_to_files, coords="all", concat_dim="t", combine="nested"
        )

        # Add lat and lon to goes_xarray
        # Ignore boundary warnings
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            goes_xarray = calc_latlon(goes_xarray)

        # Get GOES spatial resolution
        spacial_res = goes_xarray.spatial_resolution

        # Convert x and y projections to x and y projection in meters according to http://meteothink.org/examples/meteoinfolab/satellite/geos-16.html
        sat_height = goes_xarray.goes_imager_projection.attrs[
            "perspective_point_height"
        ]

        # Subset location of interest
        if "goes" in CONFIG:
            # Subset time based on user inputs
            if (
                "min_frame_index" in CONFIG["goes"]
                or "max_frame_index" in CONFIG["goes"]
            ):
                min_frame = (
                    CONFIG["goes"]["min_frame_index"]
                    if "min_frame_index" in CONFIG["goes"]
                    else 0
                )
                max_frame = (
                    CONFIG["goes"]["max_frame_index"] + 1
                    if "max_frame_index" in CONFIG["goes"]
                    else goes_xarray.dims["t"]
                )

                goes_xarray = goes_xarray.isel(
                    t=np.arange(min_frame, max_frame),
                    drop=True,
                )

            if "bounds" in CONFIG["goes"]:
                mask_lon = (goes_xarray.lon >= CONFIG["goes"]["bounds"][0]) & (
                    goes_xarray.lon <= CONFIG["goes"]["bounds"][1]
                )
                mask_lat = (goes_xarray.lat >= CONFIG["goes"]["bounds"][2]) & (
                    goes_xarray.lat <= CONFIG["goes"]["bounds"][3]
                )

                goes_xarray = goes_xarray.CMI.where(mask_lon & mask_lat, drop=True)

            else:
                goes_xarray = goes_xarray.CMI

        else:
            raise Exception("""!=====CONFIG Missing "goes" Field=====!""")

        # Replace time dimension with minutes since first time and add other x y z coords
        first_time = goes_xarray.t.values[0]
        goes_xarray = goes_xarray.assign_coords(
            t=(goes_xarray.t.values - first_time) / np.timedelta64(1, "m"),
            projection_x_coordinate=("x", goes_xarray.x.values * sat_height),
            projection_y_coordinate=("y", goes_xarray.y.values * sat_height),
            south_north=("y", np.arange(goes_xarray.shape[1])),
            west_east=("x", np.arange(goes_xarray.shape[2])),
            x=("x", np.arange(goes_xarray.shape[2])),
            y=("y", np.arange(goes_xarray.shape[1])),
        )

        # Drop incorrect projection vars
        goes_xarray = goes_xarray.drop_vars(["x_image", "y_image"])

        # Add spatial resolution info
        goes_xarray = goes_xarray.assign_attrs({"spatial_resolution": spacial_res})

        # Adjust dimension names to be standards accepted by iris
        goes_xarray["t"] = goes_xarray.t.assign_attrs(
            {
                "standard_name": "time",
                "long_name": f"minutes since {first_time}",
                "units": f"minutes since {first_time}",
            }
        )
        goes_xarray["lat"] = goes_xarray.lat.assign_attrs({"standard_name": "latitude"})
        goes_xarray["lon"] = goes_xarray.lon.assign_attrs(
            {"standard_name": "longitude"}
        )
        goes_xarray["projection_x_coordinate"] = (
            goes_xarray.projection_x_coordinate.assign_attrs({"units": "m"})
        )
        goes_xarray["projection_y_coordinate"] = (
            goes_xarray.projection_y_coordinate.assign_attrs({"units": "m"})
        )

        return goes_xarray

    raise Exception(
        f"!=====Invalid Tracking Variable. You Entered: {tracking_var.lower()}=====!"
    )



"""
<p style="text-align: center">===============Welcome To CoCoMET===============</p>

<p style="text-align: center">A toolkit of the Advanced Study of Cloud and Environment iNTerations (ASCENT) program.</p>

<p style="text-align: center">This project was supported by the U.S. Department of Energy (DOE) Early Career Research Program, Atmospheric System Research (ASR) program, and the Office of Workforce Development for Teachers and Scientists (WDTS) under the Science Undergraduate Laboratory Internships Program (SULI).</p>

<p style="text-align: center">If you are using this software for a publication, please cite: Hahn et al. (2025) https://doi.org/10.5194/egusphere-2025-1328</p>

<p style="text-align: center">=============================================</p>
"""

__all__ = [
    "user_interface_layer",
    "run_tracker_wrapper",
    "goes_load",
    "mesonh_load",
    "multi_nexrad_load",
    "nexrad_load",
    "rams_load",
    "standard_radar_load",
    "wrf_load",
    "mesonh_calculate_products",
    "rams_calculate_products",
    "wrf_calculate_products",
    "goes_tobac",
    "mesonh_tobac",
    "multi_nexrad_tobac",
    "nexrad_tobac",
    "rams_tobac",
    "standard_radar_tobac",
    "wrf_tobac",
    "mesonh_moaap",
    "rams_moaap",
    "wrf_moaap",
    "mesonh_tams",
    "rams_tams",
    "wrf_tams",
    "tracker_output_translation_layer",
    "analysis",
    "post_processor",
    "user_utils",
]

from CoCoMET.analysis.calc_var import calc_var
from CoCoMET.MOAAP import moaap
from CoCoMET.TAMS import run
from CoCoMET.user_utils import *

from .goes_load import *
from .goes_tobac import *
from .mesonh_calculate_products import *
from .mesonh_load import *
from .mesonh_moaap import *
from .mesonh_tams import *
from .mesonh_tobac import *
from .mesonhcube import *
from .multi_nexrad_load import *
from .multi_nexrad_tobac import *
from .nexrad_load import *
from .nexrad_tobac import *
from .post_processor import *
from .rams_calculate_products import *
from .rams_configure import *
from .rams_load import *
from .rams_moaap import *
from .rams_tams import *
from .rams_tobac import *
from .ramscube import *
from .run_tracker_wrapper import *
from .standard_radar_load import *
from .standard_radar_tobac import *
from .tracker_output_translation_layer import *
from .user_interface_layer import *
from .wrf_calculate_products import *
from .wrf_load import *
from .wrf_moaap import *
from .wrf_tams import *
from .wrf_tobac import *
from .wrfcube import *

print(
    "===============Welcome To CoCoMET===============\n\n"
    + "A toolkit of the Advanced Study of Cloud and Environment iNTerations (ASCENT) program.\n\n"
    + "This project was supported by the U.S. Department of Energy (DOE) Early Career Research Program, Atmospheric System Research (ASR) program, and the Office of Workforce Development for Teachers and Scientists (WDTS) under the Science Undergraduate Laboratory Internships Program (SULI).\n\n"
    + "If you are using this software for a publication, please cite: Hahn et al. (2025) https://doi.org/10.5194/egusphere-2025-1328\n\n"
    "=============================================\n"
)
